/*=============================================================================
	LaunchInitEngine.h: Unreal launcher.

Revision history:
	* Created by Sebastian Kaufel.
=============================================================================*/

//#ifdef LAUNCH_KLINGONS
//inline DOUBLE appSecondsSlow() { return appSeconds(); }
//#endif

/*-----------------------------------------------------------------------------
	InitEngine.
-----------------------------------------------------------------------------*/

static UEngine* InitEngine()
{
	guard(InitEngine);
	LAUNCH_TIME_TYPE LoadTime = appSecondsSlow();

	// Set exec hook.
	static FExecHook GLocalHook;
	GExec = &GLocalHook;

	// Create mutex so installer knows we're running.
	CreateMutexX( NULL, 0, LAUNCH_MUTEX_NAME );
#if LAUNCH_CONFIG_UI
	UBOOL AlreadyRunning = (GetLastError()==ERROR_ALREADY_EXISTS);
#endif // LAUNCH_CONFIG_UI

	// First-run menu.
	INT FirstRun=0;
	GConfig->GetInt( TEXT("FirstRun"), TEXT("FirstRun"), FirstRun );
	if( ParseParam(appCmdLine(),TEXT("FirstRun")) )
		FirstRun=0;

#ifdef LAUNCH_UNREAL_GOLD
	if( !GIsEditor && GIsClient )
	{
		if( FirstRun<ENGINE_VERSION || ParseParam(appCmdLine(),TEXT("changevideo")) )
		{
			WConfigWizard VideoDialog;
			VideoDialog.Title = LocalizeGeneral(TEXT("Video"),TEXT("Startup") );
			WWizardPage* Page = new WConfigPageRenderer( &VideoDialog );
			if( Page )
			{
				VideoDialog.Advance( Page );
				if( !VideoDialog.DoModal() )
					return NULL;
			}
		}
	}
#endif // LAUNCH_UNREAL_GOLD

#if !LAUNCH_CONFIG_UI
	// First-run and version upgrading code.
	if( FirstRun<ENGINE_VERSION && !GIsEditor && GIsClient )
	{
		// Get system directory.
		TCHAR SysDir[256]=TEXT(""), WinDir[256]=TEXT("");
#if UNICODE
		if( !GUnicodeOS )
		{
			ANSICHAR ASysDir[256]="", AWinDir[256]="";
			GetSystemDirectoryA( ASysDir, ARRAY_COUNT(SysDir) );
			GetWindowsDirectoryA( AWinDir, ARRAY_COUNT(WinDir) );
			appStrcpy( SysDir, ANSI_TO_TCHAR(ASysDir) );
			appStrcpy( WinDir, ANSI_TO_TCHAR(AWinDir) );
		}
		else
#endif
		{
			GetSystemDirectory( SysDir, ARRAY_COUNT(SysDir) );
			GetWindowsDirectory( WinDir, ARRAY_COUNT(WinDir) );
		}
#endif // !LAUNCH_CONFIG_UI
#ifndef LAUNCH_RUNE
#ifndef LAUNCH_KLINGONS
		if( FirstRun<220 )
		{
			// Migrate savegames.
			TArray<FString> Saves = GFileManager->FindFiles( TEXT("..\\Save\\*.usa"), 1, 0 );
			for( TArray<FString>::TIterator It(Saves); It; ++It )
			{
				INT Pos = appAtoi(**It+4);
				FString Section = TEXT("UnrealShare.UnrealSlotMenu");
				FString Key     = FString::Printf(TEXT("SlotNames[%i]"),Pos);
				if( appStricmp(GConfig->GetStr(*Section,*Key,TEXT("user")),TEXT(""))==0 )
					GConfig->SetString(*Section,*Key,TEXT("Saved game"),TEXT("user"));
			}
		}
#endif // !LAUNCH_KLINGONS
#endif // !LAUNCH_RUNE
#if !LAUNCH_CONFIG_UI
		if( FirstRun<220 && !GIsMMX )
		{
			// Optimize for non-MMX.
			::MessageBox( NULL, Localize(TEXT("FirstRun"),TEXT("NonMMX"),TEXT("Window")), Localize(TEXT("FirstRun"),TEXT("Caption"),TEXT("Window")), MB_OK|MB_ICONINFORMATION|MB_TASKMODAL );
			GConfig->SetString( TEXT("Galaxy.GalaxyAudioSubsystem"), TEXT("OutputRate"),      TEXT("11025Hz") );
			GConfig->SetString( TEXT("Galaxy.GalaxyAudioSubsystem"), TEXT("UseReverb"),       TEXT("False") );
			GConfig->SetString( TEXT("Galaxy.GalaxyAudioSubsystem"), TEXT("UseSpatial"),      TEXT("False") );
			GConfig->SetString( TEXT("Galaxy.GalaxyAudioSubsystem"), TEXT("UseFilter"),       TEXT("False") );
			GConfig->SetString( TEXT("Galaxy.GalaxyAudioSubsystem"), TEXT("EffectsChannels"), TEXT("8") );
		}
		if( FirstRun<220 && GPhysicalMemory <= 32*1024*1024 )
		{
			// Low memory detection.
			TCHAR Temp[4096];
			appSprintf( Temp, Localize(TEXT("FirstRun"),TEXT("LowMemory"),TEXT("Window")), GPhysicalMemory/1024/1024 );
			::MessageBox( NULL, Temp, Localize(TEXT("FirstRun"),TEXT("Caption"),TEXT("Window")), MB_OK|MB_ICONINFORMATION|MB_TASKMODAL );
			GConfig->SetBool( TEXT("Galaxy.GalaxyAudioSubsystem"), TEXT("LowSoundQuality"), 1 );
			GConfig->SetBool( TEXT("WinDrv.WindowsClient"), TEXT("LowDetailTextures"), 1 );
		}
		if( FirstRun<220 && (!GIsMMX || !GIsPentiumPro) )
		{
			// Low res detection.
			GConfig->SetString( TEXT("WinDrv.WindowsClient"), TEXT("WindowedViewportX"), TEXT("320") );
			GConfig->SetString( TEXT("WinDrv.WindowsClient"), TEXT("WindowedViewportY"), TEXT("240") );
			GConfig->SetString( TEXT("WinDrv.WindowsClient"), TEXT("WindowedColorBits"), TEXT("16") );
			GConfig->SetString( TEXT("WinDrv.WindowsClient"), TEXT("FullscreenViewportX"), TEXT("320") );
			GConfig->SetString( TEXT("WinDrv.WindowsClient"), TEXT("FullscreenViewportY"), TEXT("240") );
			GConfig->SetString( TEXT("WinDrv.WindowsClient"), TEXT("FullscreenColorBits"), TEXT("16") );
		}
#ifndef LAUNCH_UNREAL_GOLD
#ifndef LAUNCH_KLINGONS
		if( FirstRun<220 )
		{
			// Autodetect and ask about detected render devices.
			TArray<FRegistryObjectInfo> RenderDevices;
			UObject::GetRegistryObjects( RenderDevices, UClass::StaticClass(), URenderDevice::StaticClass(), 0 );
			for( INT i=0; i<RenderDevices.Num(); i++ )
			{
				TCHAR File1[256], File2[256];
				appSprintf( File1, TEXT("%s\\%s"), SysDir, *RenderDevices(i).Autodetect );
				appSprintf( File2, TEXT("%s\\%s"), WinDir, *RenderDevices(i).Autodetect );
				if( RenderDevices(i).Autodetect!=TEXT("") && (GFileManager->FileSize(File1)>=0 || GFileManager->FileSize(File2)>=0) )
				{
					TCHAR Path[256], *Str;
					appStrcpy( Path, *RenderDevices(i).Object );
					Str = appStrstr(Path,TEXT("."));
					if( Str )
					{
						*Str++ = 0;
						if( ::MessageBox( NULL, Localize(Str,TEXT("AskInstalled"),Path), Localize(TEXT("FirstRun"),TEXT("Caption"),TEXT("Window")), MB_YESNO|MB_ICONQUESTION|MB_TASKMODAL )==IDYES )
						{
							if( ::MessageBox( NULL, Localize(Str,TEXT("AskUse"),Path), Localize(TEXT("FirstRun"),TEXT("Caption"),TEXT("Window")), MB_YESNO|MB_ICONQUESTION|MB_TASKMODAL )==IDYES )
							{
								GConfig->SetString( TEXT("Engine.Engine"), TEXT("GameRenderDevice"), *RenderDevices(i).Object );
								GConfig->SetString( TEXT("SoftDrv.SoftwareRenderDevice"), TEXT("HighDetailActors"), TEXT("True") );
								break;
							}
						}
					}
				}
			}
		}
#endif  // !LAUNCH_KLINGONS
#endif  // !LAUNCH_UNREAL_GOLD
		if( FirstRun<ENGINE_VERSION )
		{
			// Initiate first run of this version.
			FirstRun = ENGINE_VERSION;
			GConfig->SetInt( TEXT("FirstRun"), TEXT("FirstRun"), FirstRun );
			::MessageBox( NULL, Localize(TEXT("FirstRun"),TEXT("Starting"),TEXT("Window")), Localize(TEXT("FirstRun"),TEXT("Caption"),TEXT("Window")), MB_OK|MB_ICONINFORMATION|MB_TASKMODAL );
		}
	}
#endif // !LAUNCH_CONFIG_UI

#if LAUNCH_CONFIG_UI
	// Commandline (for mplayer/heat)
	FString Command;
	if( Parse(appCmdLine(),TEXT("consolecommand="), Command) )
	{
		debugf(TEXT("Executing console command %s"),*Command);
		GExec->Exec( *Command, *GLog );
		return NULL;
	}

	// Test render device.
	FString Device;
	if( Parse(appCmdLine(),TEXT("testrendev="),Device) )
	{
		debugf(TEXT("Detecting %s"),*Device);
		try
		{
			UClass* Cls = LoadClass<URenderDevice>( NULL, *Device, NULL, 0, NULL );
			GConfig->SetInt(*Device,TEXT("DescFlags"),RDDESCF_Incompatible);
			GConfig->Flush(0);
			if( Cls )
			{
				URenderDevice* RenDev = ConstructObject<URenderDevice>(Cls);
				if( RenDev )
				{
					if( RenDev->Init(NULL,0,0,0,0) )
					{
						debugf(TEXT("Successfully detected %s"),*Device);
					}
					else delete RenDev;
				}
			}
		} catch( ... ) {}
		FArchive* Ar = GFileManager->CreateFileWriter(TEXT("Detected.ini"),0);
		if( Ar )
			delete Ar;
		return NULL;
	}

	// Config UI.
	guard(ConfigUI);
	if( !GIsEditor && GIsClient )
	{
		WConfigWizard D;
		WWizardPage* Page = NULL;
		if( ParseParam(appCmdLine(),TEXT("safe")) || appStrfind(appCmdLine(),TEXT("readini")) )
			{Page = new WConfigPageSafeMode(&D); D.Title=LocalizeGeneral(TEXT("SafeMode"),TEXT("Startup"));}
#ifdef LAUNCH_DEUSEX
		else if ( FirstRun<400 )
#else
		else if( FirstRun<ENGINE_VERSION )		
#endif
			{Page = new WConfigPageRenderer(&D); D.Title=LocalizeGeneral(TEXT("FirstTime"),TEXT("Startup"));}
		else if( ParseParam(appCmdLine(),TEXT("changevideo")) )
			{Page = new WConfigPageRenderer(&D); D.Title=LocalizeGeneral(TEXT("Video"),TEXT("Startup"));}
		else if( !AlreadyRunning && GFileManager->FileSize(TEXT("Running.ini"))>=0 )
			{Page = new WConfigPageSafeMode(&D); D.Title=LocalizeGeneral(TEXT("RecoveryMode"),TEXT("Startup"));}
		if( Page )
		{
			ExitSplash();
			D.Advance( Page );
			if( !D.DoModal() )
				return NULL;
			InitSplash(NULL);
		}
	}
	unguard;

	// Update first-run.
	if( FirstRun<ENGINE_VERSION )
		FirstRun = ENGINE_VERSION;
	GConfig->SetInt( TEXT("FirstRun"), TEXT("FirstRun"), FirstRun );
#endif // LAUNCH_CONFIG_UI

#if LAUNCH_CD_CHECK
	// Cd check.
	FString CdPath;
	GConfig->GetString( TEXT("Engine.Engine"), TEXT("CdPath"), CdPath );
	if
	(	CdPath!=TEXT("")
	&&	GFileManager->FileSize(TEXT("..\\Textures\\Palettes.utx"))<=0 )//oldver
	{
		FString Check = CdPath * TEXT("Textures\\Palettes.utx");
		while( !GIsEditor && GFileManager->FileSize(*Check)<=0 )
		{
			if( MessageBox
			(
				NULL,
				LocalizeGeneral("InsertCdText",TEXT("Window")),
				LocalizeGeneral("InsertCdTitle",TEXT("Window")),
				MB_TASKMODAL|MB_OKCANCEL
			)==IDCANCEL )
			{
				GIsCriticalError = 1;
				ExitProcess( 0 );
			}
		}
	}
#endif // LAUNCH_CD_CHECK

#if LAUNCH_CONFIG_UI
	// Create is-running semaphore file.
	FArchive* Ar = GFileManager->CreateFileWriter(TEXT("Running.ini"),0);
	if( Ar )
		delete Ar;
#endif // LAUNCH_CONFIG_UI

#if ENGINE_VERSION<230 && !LAUNCH_KLINGONS
	// Display the damn story to appease the German censors.
	UBOOL CanModifyGore=1;
	GConfig->GetBool( TEXT("UnrealI.UnrealGameOptionsMenu"), TEXT("bCanModifyGore"), CanModifyGore );
	if( !CanModifyGore && !GIsEditor )
	{
		FString S;
		if( appLoadFileToString( S, TEXT("Story.txt") ) )
		{
			WTextScrollerDialog Dlg( TEXT("The Story"), *S );
			Dlg.DoModal();
		}
	}
#endif

#ifdef LAUNCH_RUNE
	// RUNE: Display ESRB message
	if ( FirstRun==0 && !GIsEditor )
	{
		FString S;
		if( appLoadFileToString( S, TEXT("License.txt") ) )
		{
			ExitSplash();
			WTextScrollerDialog Dlg( TEXT("License"), *S );
			Dlg.DoModal();
			InitSplash(NULL);
		}
	}
#endif

	// Create the global engine object.
	UClass* EngineClass;
	if( !GIsEditor )
	{
		// Create game engine.
		EngineClass = UObject::StaticLoadClass( UGameEngine::StaticClass(), NULL, TEXT("ini:Engine.Engine.GameEngine"), NULL, LOAD_NoFail, NULL );
	}
	else
	{
		// Editor.
		EngineClass = UObject::StaticLoadClass( UEngine::StaticClass(), NULL, TEXT("ini:Engine.Engine.EditorEngine"), NULL, LOAD_NoFail, NULL );
	}
	UEngine* Engine = ConstructObject<UEngine>( EngineClass );
	Engine->Init();

	debugf( TEXT("Startup time: %f seconds"), appSecondsSlow()-LoadTime );

	return Engine;
	unguard;
}

/*-----------------------------------------------------------------------------
	The End.
-----------------------------------------------------------------------------*/
